;;;============================================================================

;;; File: "rm.scm", Time-stamp: <2008-11-08 09:01:37 feeley>

;;; Copyright (c) 2008 by Marc Feeley, All Rights Reserved.

;;;============================================================================

(define rm-default-server-address #f)    ;; allow connections on all interfaces
(define rm-default-server-port-num 8000)

(define (rm-split-address address)
  (let* ((x
          (if address
              (call-with-input-string
               address
               (lambda (port)
                 (read-all port (lambda (port) (read-line port #\:)))))
              '()))
         (len
          (length x)))
    (cond ((<= len 1)
           (cons (if (= len 1)
                     (car x)
                     rm-default-server-address)
                 rm-default-server-port-num))
          ((= len 2)
           (let ((server-address (car x))
                 (server-port-num (string->number (cadr x) 10)))
             (if (and server-port-num
                      (exact? server-port-num)
                      (integer? server-port-num)
                      (>= server-port-num 1)
                      (<= server-port-num 65535))
                 (cons (if (string=? server-address "")
                           rm-default-server-address
                           server-address)
                       server-port-num)
                 #f)))
          (else
           #f))))

;;;----------------------------------------------------------------------------

(define-type rm-state
  constructor: rm-make-state
  random-source
)

(define rm-state #f)

(define (rm-setup)
  (let ((rs (make-random-source)))
    (random-source-randomize! rs)
    (set! rm-state
          (rm-make-state rs))))

(define (rm-make-unique-id)
  (vector
   ((random-source-make-integers (rm-state-random-source rm-state))
    1000000000000000000000000000000000000)))

(define (rm-cleanup)
  (set! rm-state #f))

;;;----------------------------------------------------------------------------

(define-type rm-ctx
  constructor: rm-make-ctx
  id
  server-address-and-port-num
  function-mapper
  connection-ctx-table
)

(define (rm-ctx-create)
  (let ((ctx
         (rm-make-ctx
          (rm-make-unique-id)
          #f
          #f
          (make-table weak-keys: #t))))
    (rm-server-configure ctx #f)
    ctx))

(define (rm-server-configure ctx address)
  (let ((server-address-and-port-num (rm-split-address address)))
    (rm-ctx-server-address-and-port-num-set!
     ctx
     server-address-and-port-num)
    ctx))

(define (rm-server-start ctx function-mapper)
  (rm-ctx-function-mapper-set! ctx function-mapper)
  (rm-server-register ctx))

(define (rm-server-register ctx)
  (let ((server-address
         (car (rm-ctx-server-address-and-port-num ctx)))
        (port-num
         (cdr (rm-ctx-server-address-and-port-num ctx))))
    (tcp-service-register!
     (list server-address: server-address
           port-number: port-num)
     (lambda ()
       (rm-serve-client ctx)))))

(define (rm-connection-ctx-register ctx connection-ctx)
  (table-set!
   (rm-ctx-connection-ctx-table ctx)
   (rm-connection-ctx-id connection-ctx)
   connection-ctx))

;;;----------------------------------------------------------------------------

(define-type rm-connection-ctx
  constructor: rm-make-connection-ctx
  id
  input-port
  output-port
  reader-thread
  writer-thread
  function-mapper
  seqnum
  call-table
)

(define (rm-serve-client ctx)
  (let ((input-port
         (current-input-port))
        (output-port
         (current-output-port)))
    (rm-input-serialization-setup input-port)
    (rm-output-serialization-setup output-port)
    (let ((id (rm-read-ctx-id input-port)))
      (if id
          (begin
            (rm-write-ctx-id output-port ctx)
            (let ((connection-ctx
                   (rm-make-connection-ctx
                    id
                    input-port
                    output-port
                    #f
                    #f
                    (rm-ctx-function-mapper ctx)
                    0
                    (make-table))))
              (let ((reader-thread
                     (rm-create-reader-thread connection-ctx))
                    (writer-thread
                     (current-thread)))
                (rm-connection-ctx-reader-thread-set!
                 connection-ctx
                 reader-thread)
                (rm-connection-ctx-writer-thread-set!
                 connection-ctx
                 writer-thread)
                (rm-connection-ctx-register ctx connection-ctx)
                (thread-start! reader-thread)
                (rm-writer connection-ctx))))))))

(define (rm-connect ctx address function-mapper)
  (let* ((server-address-and-port-num
          (rm-split-address address))
         (server-address
          (car server-address-and-port-num))
         (port-num
          (cdr server-address-and-port-num)))
    (let ((connection
           (open-tcp-client
            (list server-address: server-address
                  port-number: port-num))))
      (rm-input-serialization-setup connection)
      (rm-output-serialization-setup connection)
      (rm-write-ctx-id connection ctx)
      (let ((id (rm-read-ctx-id connection)))
        (if id
            (let ((connection-ctx
                   (rm-make-connection-ctx
                    id
                    connection
                    connection
                    #f
                    #f
                    function-mapper
                    0
                    (make-table))))
              (let ((reader-thread
                     (rm-create-reader-thread connection-ctx))
                    (writer-thread
                     (rm-create-writer-thread connection-ctx)))
                (rm-connection-ctx-reader-thread-set!
                 connection-ctx
                 reader-thread)
                (rm-connection-ctx-writer-thread-set!
                 connection-ctx
                 writer-thread)
                (rm-connection-ctx-register ctx connection-ctx)
                (thread-start! reader-thread)
                (thread-start! writer-thread)
                connection-ctx))
            #f)))))

(define (rm-read-ctx-id input-port)
  (let ((head (read input-port)))
    (and (pair? head)
         (pair? (cdr head))
         (pair? (cddr head))
         (null? (cdddr head))
         (eq? (car head) 'gambit-rm)
         (eqv? (cadr head) 0)
         (caddr head))))

(define (rm-write-ctx-id output-port ctx)
  (write (list 'gambit-rm 0 (rm-ctx-id ctx)) output-port)
  (force-output output-port))

(define (rm-recv connection-ctx)
  (let ((in (rm-connection-ctx-output-port connection-ctx)))
    (rm-import connection-ctx (read in))))

(define (rm-send connection-ctx msg)
  (let ((out (rm-connection-ctx-output-port connection-ctx)))
    (write (rm-export connection-ctx msg) out)
    (force-output out)))

(define (rm-input-serialization-setup port)
  (input-port-readtable-set!
   port
   (readtable-sharing-allowed?-set
    (input-port-readtable port)
    'serialize)))

(define (rm-output-serialization-setup port)
  (output-port-readtable-set!
   port
   (readtable-sharing-allowed?-set
    (output-port-readtable port)
    'serialize)))

(define (rm-serialization-setup connection-ctx)
  (rm-input-serialization-setup
   (rm-connection-ctx-input-port connection-ctx))
  (rm-output-serialization-setup
   (rm-connection-ctx-output-port connection-ctx)))

;;;----------------------------------------------------------------------------

(define (rm-create-reader-thread connection-ctx)
  (make-thread (lambda () (rm-reader connection-ctx))))

(define (rm-reader connection-ctx)
  (let loop ()
    (let ((msg (rm-recv connection-ctx)))
      (if (not (eof-object? msg))
          (begin
            (thread-send (rm-connection-ctx-writer-thread connection-ctx)
                         msg)
            (loop)))))
  (thread-send (rm-connection-ctx-writer-thread connection-ctx)
               '(reader-terminated)))

(define (rm-create-writer-thread connection-ctx)
  (make-thread (lambda () (rm-writer connection-ctx))))

(define (rm-writer connection-ctx)
  (let loop ()
    (let ((msg (thread-receive)))
      (and (rm-process-message connection-ctx msg)
           (loop)))))

(define (rm-new-seqnum connection-ctx)
  (let ((seqnum (+ (rm-connection-ctx-seqnum connection-ctx) 1)))
    (rm-connection-ctx-seqnum-set! connection-ctx seqnum)
    seqnum))

(define (rm-process-call connection-ctx call)
  (apply ((rm-connection-ctx-function-mapper connection-ctx)
          connection-ctx
          (car call))
         (cons connection-ctx
               (cdr call))))

(define (rm-process-message connection-ctx msg)
  (and (vector? msg)

       (case (vector-ref msg 0)

         ((terminate)
          #f)

         ((remote-call)
          (let* ((result-mutex
                  (vector-ref msg 1))
                 (call
                  (vector-ref msg 2))
                 (seqnum
                  (rm-new-seqnum connection-ctx)))
            (table-set! (rm-connection-ctx-call-table connection-ctx)
                        seqnum
                        result-mutex)
            (rm-send connection-ctx
                     (vector 'call seqnum call))
            #t))

         ((call)
          (let* ((seqnum
                  (vector-ref msg 1))
                 (call
                  (vector-ref msg 2))
                 (result
                  (rm-process-call connection-ctx call)))
            (rm-send connection-ctx
                     (vector 'return seqnum result))
            #t))

         ((return)
          (let* ((seqnum
                  (vector-ref msg 1))
                 (result
                  (vector-ref msg 2))
                 (call-table
                  (rm-connection-ctx-call-table connection-ctx))
                 (result-mutex
                  (table-ref call-table seqnum #f)))
            (if (not result-mutex)
                (error "invalid call sequence number")
                (begin
                  (table-set! call-table seqnum) ;; remove call
                  (mutex-specific-set! result-mutex result)
                  (mutex-unlock! result-mutex)
                  #t))))

         (else
          (pretty-print
           (list 'unhandled-message msg)
           ##stdout-port);;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
          #t))))

(define (rm-remote-call connection-ctx fn . args)
  (let ((result-mutex (make-mutex 'remote-call)))
    (mutex-lock! result-mutex) ;; result not ready yet...
    (thread-send
     (rm-connection-ctx-writer-thread connection-ctx)
     (vector 'remote-call result-mutex (cons fn args)))
    (mutex-lock! result-mutex) ;; wait until result is ready
    (mutex-specific result-mutex)))

(define (rm-export connection-ctx obj)
  (object->$object connection-ctx obj))

(define (rm-import connection-ctx obj)
  ($object->object connection-ctx obj))

;;;-----------------------------------------------------------------------------

(define-type $object
  id: cc36b50e-dde5-4716-a65f-c63607346345
  extender: define-type-of-object
  connection-ctx-id
)

(define (object->$object connection-ctx obj)
  (cond ((thread? obj)
         (thread->$thread connection-ctx obj))
        ((pair? obj)
         (cons (object->$object connection-ctx (car obj))
               (object->$object connection-ctx (cdr obj))))
        ((vector? obj)
         (list->vector
          (map (lambda (x) (object->$object connection-ctx x))
               (vector->list obj))))
        ((or (number? obj)
             (string? obj)
             (symbol? obj)
             (boolean? obj)
             (null? obj)
             (eof-object? obj)
             ($thread? obj))
         obj)
        (else
         (error "object->$object can't handle" obj))))
        
(define ($object->object connection-ctx $obj)
  (cond (($thread? $obj)
         ($thread->thread connection-ctx $obj))
        ((pair? $obj)
         (cons ($object->object connection-ctx (car $obj))
               ($object->object connection-ctx (cdr $obj))))
        ((vector? $obj)
         (list->vector
          (map (lambda (x) ($object->object connection-ctx x))
               (vector->list $obj))))
        ((or (number? $obj)
             (string? $obj)
             (symbol? $obj)
             (boolean? $obj)
             (null? $obj)
             (eof-object? $obj))
         $obj)
        (else
         (error "$object->object can't handle" $obj))))
        
;;;-----------------------------------------------------------------------------

(define-type-of-object $thread
  id: ad057bf6-e636-4ebb-84e7-d3360cf9f618
  sn
)

(define (thread->$thread connection-ctx t)
  (let ((id (rm-connection-ctx-id connection-ctx)))
    (make-$thread id
                  (object->serial-number t))))

(define ($thread->thread connection-ctx t)
  (let ((id (rm-connection-ctx-id connection-ctx)))
    (if (equal? ($object-connection-ctx-id t) id)
        (serial-number->object ($thread-sn t))
        t)))

;;;============================================================================
